(* This Mathematica script 
  computes the linear Galerkin FEM stiffness matrix on a given mesh,  
  by OSC-like relations; see (bibtex entry)
  author = 	 {Mario Putti and Christian Cordes},
  title = 	 {Finite Element Approximation of the Diffusion
		  Operator on Tetrahedra},
  journal = 	 {SIAM J. Sci. Stat. Comput.},
  year = 	 1998,
  volume =       {19},
  number =       {4},
  pages =        {1154--1168".

Afterwards, it builds the FEM linear system *)
  
(* Requires: << readMesh.m *)

stiff = Table[ 0, {i,1,nnod}, {j,1,nnod} ];
(* consider angles corresponding to each vertex of each triangle *)
tour = {{2,1,3},{1,2,3},{1,3,2}}

For[ itri = 1, itri <= ntri, itri++,
 sum = 0;
 i1 = inodtri[[ itri, 1 ]];
 i2 = inodtri[[ itri, 2 ]];
 i3 = inodtri[[ itri, 3 ]];
 x1 = xv[[ i1 ]];
 y1 = yv[[ i1 ]];
 x2 = xv[[ i2 ]];
 y2 = yv[[ i2 ]];
 x3 = xv[[ i3 ]];
 y3 = yv[[ i3 ]];
 area2T = Abs[ Det[ {{x1,y1,1},{x2,y2,1},{x3,y3,1}} ] ];
 (* Print["tri=",itri,"; area2T=",N[area2T] ]; *)
 For [ iang=1, iang <= 3, iang++,
  {ind1,inda,ind2} = tour[[ iang ]];
  (* compute the tangent of the angle corresp. to vertex ia *) 
  ia = inodtri[[ itri, inda ]];
  i1 = inodtri[[ itri, ind1 ]];
  i2 = inodtri[[ itri, ind2 ]];
  xa = xv[[ ia ]];
  ya = yv[[ ia ]];
  x1 = xv[[ i1 ]];
  y1 = yv[[ i1 ]];
  x2 = xv[[ i2 ]];
  y2 = yv[[ i2 ]];
  lenPaP1 = Sqrt[ (x1 - xa)^2 + (y1 - ya)^2 ];
  lenPaP2 = Sqrt[ (x2 - xa)^2 + (y2 - ya)^2 ];
  lenP1P2 = Sqrt[ (x1 - x2)^2 + (y1 - y2)^2 ];
  (* compute height w.r.t. Pa-P1 *)
  lenh = area2T / lenPaP1;
  (* compute projection of Pa-P2 on Pa-P1 *)
  lenb = Sqrt[ lenPaP2^2 - lenh^2 ];
  If [ (lenb == 0), abstan = Infinity, abstan = lenh / lenb];
  (* compute sign *)
  cosang = {x1-xa,y1-ya} . {x2-xa,y2-ya} / (lenPaP1 * lenPaP2);
  (* our angles are < Pi, hence ... *)
  If[ (cosang == 0), signtan = 1, signtan = Sign[ cosang ] ];
  If[ (abstan != Infinity),
   tan = signtan * abstan;
   atan = ArcTan[ tan ],
  (* else *)
   tan = Infinity;
   atan = Pi / 2
  ];
  If [ (atan < 0), atan1 = atan + Pi, atan1 = atan ];
  sum = sum + atan1;
  (* Print["tri=",itri,"; angle for node=",ia,"; tan=",N[tan],
	   "; atan1=",N[atan1]]; *)
  If[ (tan == 0), Print["ERROR: tan=",N[tan]]; Abort[] ]; 
  If[ (tan == Infinity), sij = 0, sij = - 1 / (2 * tan) ]; 
  stiff[[ i1, i2 ]] = stiff[[ i1, i2 ]] + sij;
  stiff[[ i2, i1 ]] = stiff[[ i2, i1 ]] + sij;
     (* If[ ((i1 == 6) || (i2 == 6)), Dialog[] ]; *)
 ]; (* end For iang *)
 (* Print["tri=",itri,"; sum angles=",N[sum] ]; *)
] (* end For itri *)

(* compute diagonal elements, by recalling that (sum_j s_ij = 0) *)
For[ inod = 1, inod <= nnod, inod++,
 sum = 0;
 For[ jnod = 1, jnod <= nnod, jnod++,
  sum = sum + stiff[[ inod, jnod ]];
 ];
 stiff[[ inod, inod ]] = stiff[[ inod, inod ]] - sum; 
]

(* compute and write diagonally preconditioned stiffness matrix *)
stiffPrec = stiff;
For[ inod = 1, inod <= nnod, inod++,
 diagEl = stiff[[ inod, inod ]];
 For[ jnod = 1, jnod <= nnod, jnod++,
  stiffPrec[[ inod, jnod ]] = stiffPrec[[ inod, jnod ]] / diagEl;
 ];
]

Export["stiffPrec.csv",N[stiffPrec],"CSV"]

utarrT = Array[ ut, nnod ]

eqs = {}

For[ inod = 1, inod <= nnod, inod++,
 lhs = 0;
 rhs = 0;
 For[ jnod = 1, jnod <= nnod, jnod++,
  lhs = lhs + stiff[[ inod, jnod ]] * ut[jnod];
  rhs = 0;
 ];
 eqs = Append[ eqs, lhs == rhs ];
]

Print["Stiffness matrix:",stiff//MatrixForm]

Print["Diagonally preconditioned stiffness matrix:",stiffPrec//MatrixForm]
